﻿using System;
using System.Collections.Generic;
using System.IdentityModel.Tokens.Jwt;
using System.IO;
using System.Linq;
using System.Net;
using System.Runtime.InteropServices;
using System.Threading;
using System.Windows.Forms;

using Microsoft.Azure.Management.Logic;
using Microsoft.Azure.Management.Logic.Models;
using Microsoft.IdentityModel.Clients.ActiveDirectory;
using Microsoft.Rest;
using Microsoft.Rest.Azure;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.Subscriptions;
using Microsoft.WindowsAzure.Subscriptions.Models;

namespace DemoGetSubscriptionsAndLogicApps
{
    public partial class MainForm : Form
    {
        Dictionary<string, AuthenticationResult> credentialsList = new Dictionary<string, AuthenticationResult>();
        Dictionary<string, AuthenticationContext> authenticationContextList = new Dictionary<string, AuthenticationContext>();
        AuthenticationResult commonTenantCredentials = null;
        static PromptBehavior promptBehavior = PromptBehavior.RefreshSession;

        static string TokenAudience = "https://management.core.windows.net/";
        static string ClientRedirectUri = "urn:ietf:wg:oauth:2.0:oob";
        static string ClientId = "1950a258-227b-4e31-a9cf-717495945fc2";
        static string EnableMagicCookie = "site_id=501358&display=popup";

        const string LoginUrlBase = "https://login.windows.net/{0}";
        const string LogoutUrlBase = "https://login.windows.net/{0}/oauth2/logout";

        public MainForm()
        {
            InitializeComponent();
        }
        
        private void btnLogin_Click(object sender, EventArgs e)
        {
            labAccountAdminValue.Text = "n/a";
            labSubAdminValue.Text = "n/a";
            labSubIdValue.Text = "n/a";
            labSubNameValue.Text = "n/a";
            labTenantIdValue.Text = "n/a";
            labSubStatusValue.Text = "n/a";
            btnListLogicApps.Enabled = false;
            lstLogicApps.Enabled = false;
            lstLogicApps.Items.Clear();
            credentialsList.Clear();

            commonTenantCredentials = GetAuthorizationToken();

            if (commonTenantCredentials != null)
            {
                // Add to dictionary, keyed against tenant id
                credentialsList.Add("common", commonTenantCredentials);

                // Get JWT token
                JwtSecurityTokenHandler jwtHandler = new JwtSecurityTokenHandler();
                JwtSecurityToken secToken = jwtHandler.ReadToken(commonTenantCredentials.AccessToken) as JwtSecurityToken;

                string userName;
                string orgName;

                string[] uniqueNameParts = secToken.Payload["unique_name"].ToString().Split('#');
                if (uniqueNameParts.Length == 1)
                {
                    // Org id
                    userName = uniqueNameParts[0];
                    orgName = uniqueNameParts[0].Split('@')[1];
                }
                else
                {
                    // Microsoft account
                    userName = uniqueNameParts[1];
                    orgName = uniqueNameParts[0] == "live.com" ? "Microsoft Account" : uniqueNameParts[0];
                }

                labLoggedInAsValue.Text = string.Format("{0} ({1})", userName, orgName);
                labLoginValidTillValue.Text = secToken.ValidTo.ToString("G");

                promptBehavior = PromptBehavior.RefreshSession;
                btnLogin.Enabled = false;
                btnLogoff.Enabled = true;

                GetSubscriptions();
            }
        }

        private void btnLogoff_Click(object sender, EventArgs e)
        {
            labAccountAdminValue.Text = "n/a";
            labSubAdminValue.Text = "n/a";
            labSubIdValue.Text = "n/a";
            labSubNameValue.Text = "n/a";
            labTenantIdValue.Text = "n/a";
            labSubStatusValue.Text = "n/a";
            btnListLogicApps.Enabled = false;
            lstLogicApps.Enabled = false;
            lstLogicApps.Items.Clear();
            lstSubscriptions.Enabled = false;
            lstSubscriptions.Items.Clear();
            credentialsList.Clear();

            btnLogin.Enabled = false;
            btnLogoff.Enabled = false;

            WebClient client = new WebClient();

            // Clear token caches
            foreach (KeyValuePair<string, AuthenticationContext> kvp in authenticationContextList)
            {
                kvp.Value.TokenCache.Clear();

                // Logout of each tenant
                Uri logoutUri = new Uri(string.Format(LogoutUrlBase, kvp.Key));

                try
                {
                    using (Stream response = client.OpenRead(logoutUri))
                    {
                        labLoggedInAsValue.Text = "n/a";
                        labLoginValidTillValue.Text = "n/a";
                        btnLogoff.Enabled = false;
                        btnLogin.Enabled = true;
                        promptBehavior = PromptBehavior.Always;
                    }
                }
                catch (Exception)
                {
                    // timeout. That's expected
                }

            }

            authenticationContextList.Clear();
            credentialsList.Clear();

            // Clear cookies
            ClearCookies();
        }

        private void lstSubscriptions_SelectedIndexChanged(object sender, EventArgs e)
        {
            // Display the data for the selected subscription
            if (lstSubscriptions.SelectedItem != null && lstSubscriptions.SelectedItem is SubscriptionWrapper)
            {
                SubscriptionWrapper subscription = (SubscriptionWrapper)lstSubscriptions.SelectedItem;

                labAccountAdminValue.Text = subscription.AccountAdminLiveEmailId;
                labSubAdminValue.Text = subscription.ServiceAdminLiveEmailId;
                labSubIdValue.Text = subscription.SubscriptionId;
                labSubNameValue.Text = subscription.SubscriptionName;
                labTenantIdValue.Text = subscription.ActiveDirectoryTenantId;
                labSubStatusValue.Text = subscription.SubscriptionStatus;
                lstLogicApps.Enabled = false;
                lstLogicApps.Items.Clear();
                btnListLogicApps.Enabled = true;
            }
        }

        private void btnListLogicApps_Click(object sender, EventArgs e)
        {
            lstLogicApps.Items.Clear();
            lstLogicApps.Enabled = false;

            // Get the list of logic apps for the current subscription
            if (lstSubscriptions.SelectedItem != null && lstSubscriptions.SelectedItem is SubscriptionWrapper)
            {
                SubscriptionWrapper subscription = (SubscriptionWrapper)lstSubscriptions.SelectedItem;
                AuthenticationResult credentials = null;

                // Check if we already have a credentials object for this tenant
                if (credentialsList.ContainsKey(subscription.ActiveDirectoryTenantId))
                {
                    credentials = credentialsList[subscription.ActiveDirectoryTenantId];
                }
                else
                {
                    // Get the credentials for the active subscription tenant id
                    promptBehavior = PromptBehavior.Auto;
                    credentials = GetAuthorizationToken(subscription.ActiveDirectoryTenantId);
                    credentialsList[subscription.ActiveDirectoryTenantId] = credentials;
                    promptBehavior = PromptBehavior.RefreshSession;
                }

                using (LogicManagementClient client = new LogicManagementClient(new TokenCredentials(credentials.AccessToken)))
                {
                    client.SubscriptionId = subscription.SubscriptionId;
                    Page<Workflow> workflowPage = client.Workflows.ListBySubscription();

                    List<Workflow> workflows = workflowPage.ToList();

                    if (workflows.Count > 0)
                    {
                        lstLogicApps.Enabled = true;
                        foreach (Workflow wf in workflows)
                        {
                            lstLogicApps.Items.Add(string.Format("{0} ({1})", wf.Name, wf.State));
                        }
                    }
                    else
                    {
                        lstLogicApps.Items.Add("(no workflows exist in this subscription)");
                    }                    
                }
            }
        }

        private async void GetSubscriptions()
        {
            CancellationToken canToken = new CancellationToken();
            TokenCloudCredentials cred = new TokenCloudCredentials(commonTenantCredentials.AccessToken);
            SubscriptionClient client = CloudContext.Clients.CreateSubscriptionClient(cred);

            SubscriptionListOperationResponse response = await client.Subscriptions.ListAsync(canToken);

            // Project subscriptions into a list of SubscriptionWrapper instances
            List<SubscriptionWrapper> subscriptions = response.Subscriptions.Select
                (
                    sub =>
                        new SubscriptionWrapper(sub)
                ).ToList();

            lstSubscriptions.Items.Clear();
            lstSubscriptions.Items.AddRange(subscriptions.ToArray());

            lstSubscriptions.Enabled = true;
        }

        private AuthenticationResult GetRefreshToken(AuthenticationResult currentCredentials, string tenantId = "common")
        {
            AuthenticationContext authenticationContext = null;

            // Check if we have an AutenticationContext for this tenant
            if (authenticationContextList.ContainsKey(tenantId))
            {
                authenticationContext = authenticationContextList[tenantId];
            }
            else
            {
                authenticationContext = new AuthenticationContext(string.Format(LoginUrlBase, tenantId), true, TokenCache.DefaultShared);
                authenticationContextList[tenantId] = authenticationContext;
            }

            AuthenticationResult result = null;

            try
            {
                result = authenticationContext.AcquireTokenByRefreshToken(
                            currentCredentials.RefreshToken,
                            MainForm.ClientId);

            }
            catch (AdalException)
            {
                result = null;
            }
            return result;
        }

        /// <summary>
        /// Displays the Azure login window, prompts the user to login (if required), 
        /// and returns a <see cref="AuthenticationResult"/> instance for the logged-in user.
        /// </summary>
        /// <returns><see cref="AuthenticationResult"/> instance.</returns>
        private AuthenticationResult GetAuthorizationToken(string tenantId = "common")
        {
            AuthenticationContext authenticationContext = null;

            // Check if we have an AutenticationContext for this tenant
            if (authenticationContextList.ContainsKey(tenantId))
            {
                authenticationContext = authenticationContextList[tenantId];
            }
            else
            {
                authenticationContext = new AuthenticationContext(string.Format(LoginUrlBase, tenantId), true, TokenCache.DefaultShared);
                authenticationContextList[tenantId] = authenticationContext;
            }

            AuthenticationResult result = null;

            try
            {
                result = authenticationContext.AcquireToken(
                            MainForm.TokenAudience,
                            MainForm.ClientId,
                            new Uri(MainForm.ClientRedirectUri),
                            promptBehavior,
                            UserIdentifier.AnyUser,
                            MainForm.EnableMagicCookie);

            }
            catch (AdalException)
            {
                result = null;
            }
            return result;
        }

        private void ClearCookies()
        {
            NativeMethods.InternetSetOption(IntPtr.Zero, NativeMethods.INTERNET_OPTION_END_BROWSER_SESSION, IntPtr.Zero, 0);
        }

        private static class NativeMethods
        {
            internal const int INTERNET_OPTION_END_BROWSER_SESSION = 42;

            [DllImport("wininet.dll", SetLastError = true)]
            internal static extern bool InternetSetOption(IntPtr hInternet, int dwOption, IntPtr lpBuffer,
                int lpdwBufferLength);
        }
    }
}
